/*

Copyright (c) 2005-2015, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "APSimulator.hpp"

#include "RegularStimulus.hpp"
#include "SimpleStimulus.hpp"
#include "MultiStimulus.hpp"
#include "AbstractIvpOdeSolver.hpp"
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/tokenizer.hpp>
#include "BoostFilesystem.hpp"
#include "RandomNumberGenerator.hpp"

#include "CheckpointArchiveTypes.hpp"
#include "ArchiveLocationInfo.hpp"

// Include all hpp files of the models we want the option of simulating
#include "hodgkin_huxley_squid_axon_model_1952_modifiedCvodeOpt.hpp"
#include "noble_model_1962CvodeOpt.hpp"
//#include "beeler_reuter_model_1977CvodeOpt.hpp" // The below data clamp versions are equivalent to these when the data clamp current conductance is 0
//#include "luo_rudy_1991CvodeOpt.hpp"
//#include "ten_tusscher_model_2004_epiCvodeOpt.hpp"
//#include "ohara_rudy_2011_endoCvodeOpt.hpp"

#include "beeler_reuter_model_1977_data_clampCvodeOpt.hpp"
#include "luo_rudy_1991_data_clampCvodeOpt.hpp"
#include "ten_tusscher_model_2004_epi_data_clampCvodeOpt.hpp"
#include "ohara_rudy_2011_endo_data_clampCvodeOpt.hpp"
#include "davies_isap_2012CvodeDataClampOpt.hpp"

APSimulator::APSimulator(unsigned rModelChoice, unsigned protocolChoice, unsigned rSeed, int rFreezingIndex, unsigned dog)
 : mSigma(DOUBLE_UNSET), mRunningToSteadyState(false)
{
    mModelChoice = rModelChoice;
    mProtocolChoice = protocolChoice;
    mSeed = rSeed;
    mFreezingIndex = rFreezingIndex;
    mDog = dog;

    // check validity of model and protocol choices
    assert( ( mModelChoice >= 1 ) && ( mModelChoice <= 7 ) );
    assert( ( mProtocolChoice >= 1 ) && ( mProtocolChoice <= 8 ) );

    // redirecting stderr to monitor progress of MCMC
    std::string log_file_directory;
    if (mDog>0)
    {
        log_file_directory = "projects/Jmcc_UQ/output/stderr/dog_trace/";
    }
    else
    {
        log_file_directory = "projects/Jmcc_UQ/output/stderr/synthetic_data/";
    }
    boost::filesystem::create_directories(log_file_directory);

    std::string log_file_path = log_file_directory + "model_" + boost::lexical_cast<std::string>(mModelChoice);
    log_file_path += "_protocol_";
    log_file_path += boost::lexical_cast<std::string>(mProtocolChoice);
    if (mDog > 0)
    {
        log_file_path += ("_dog_"+boost::lexical_cast<std::string>(mDog));
    }
    log_file_path += "_stderr.log";

    // This solves a problem we found with stderr output "filling up" causing the MCMC to hang indefinitely
    freopen( log_file_path.c_str(), "w", stderr );

    double stimulus_duration;
    double stimulus_magnitude;
    double stimulus_period;
    double stimulus_start_time;

    if (mDog <= 0)
    {
        stimulus_duration = 2.0;
        stimulus_magnitude = -25.5;
        stimulus_period = 1000;
        stimulus_start_time = 0;
    }
    else
    {
        stimulus_duration = 2.0;
        stimulus_magnitude = 0; // With dog traces, we "data clamp" to the segment of the trace where the stimulus current is being applied, and apply no actual stimulus current to the model
        stimulus_period = 1000;
        stimulus_start_time = 0;


        mDataClampOn = 9.625;
        mDataClampOff = 11.875;
    }

    boost::shared_ptr<AbstractStimulusFunction> p_stimulus;

    if ( ( mProtocolChoice != 4u ) ) // regular pacing at single frequency for all protocols except for 4
    {
        p_stimulus = boost::shared_ptr<AbstractStimulusFunction>(new RegularStimulus(stimulus_magnitude,stimulus_duration,stimulus_period,stimulus_start_time));
    }
    else
    {
        boost::shared_ptr<RegularStimulus> p_stim1(new RegularStimulus(stimulus_magnitude,stimulus_duration,1000,0));
        boost::shared_ptr<SimpleStimulus> p_stim2(new SimpleStimulus(stimulus_magnitude, stimulus_duration, 1500));
        boost::shared_ptr<MultiStimulus> p_multi_stim(new MultiStimulus());
        p_multi_stim->AddStimulus(p_stim1);
        p_multi_stim->AddStimulus(p_stim2);
        p_stimulus = boost::shared_ptr<AbstractStimulusFunction>(p_multi_stim);
    }



    boost::shared_ptr<AbstractIvpOdeSolver> p_solver;

    // define array of parameter names to later be set to proposed values during MCMC
    if (mModelChoice==1u)
    {
        mpModel.reset(new Cellhodgkin_huxley_squid_axon_model_1952_modifiedFromCellMLCvodeOpt(p_solver, p_stimulus));
        mParameterMetanames.push_back("membrane_fast_sodium_current_conductance");  // 120
        mParameterMetanames.push_back("membrane_potassium_current_conductance");    // 36
        mParameterMetanames.push_back("leakage_current_conductance");               // 0.3
    }
    if (mModelChoice==2u)
    {
        mpModel.reset(new Cellnoble_model_1962FromCellMLCvodeOpt(p_solver, p_stimulus));
        mParameterMetanames.push_back("membrane_fast_sodium_current_conductance");                // 400
        mParameterMetanames.push_back("membrane_inward_rectifier_potassium_current_conductance");   // 0.015
        mParameterMetanames.push_back("membrane_delayed_rectifier_potassium_current_conductance");  // 1.2
        mParameterMetanames.push_back("leakage_current_conductance");                    // 0.075
    }
    else if (mModelChoice==3u)
    {
        mpModel.reset(new Cellbeeler_reuter_model_1977_data_clampFromCellMLCvodeOpt(p_solver, p_stimulus));
        mParameterMetanames.push_back("membrane_fast_sodium_current_conductance");                        // 4e-2
        mParameterMetanames.push_back("membrane_inward_rectifier_potassium_current_conductance");         // 0.0035
        mParameterMetanames.push_back("membrane_rapid_delayed_rectifier_potassium_current_conductance");  // 0.008
        mParameterMetanames.push_back("membrane_L_type_calcium_current_conductance");                     // 9e-4
    }
    else if (mModelChoice==4u)
    {
        mpModel.reset(new Cellluo_rudy_1991_data_clampFromCellMLCvodeOpt(p_solver, p_stimulus));
        mParameterMetanames.push_back("membrane_fast_sodium_current_conductance");                  // 23
        mParameterMetanames.push_back("membrane_delayed_rectifier_potassium_current_conductance");  // 0.282
        mParameterMetanames.push_back("membrane_inward_rectifier_potassium_current_conductance");   // 0.6047
        mParameterMetanames.push_back("membrane_L_type_calcium_current_conductance");                  // 0.09
        mParameterMetanames.push_back("leakage_current_conductance");                                  // 0.03921
        mParameterMetanames.push_back("membrane_plateau_potassium_current_conductance");               // 0.0183
    }
    else if (mModelChoice==5u)
    {
        mpModel.reset(new Cellten_tusscher_model_2004_epi_data_clampFromCellMLCvodeOpt(p_solver, p_stimulus));
        mParameterMetanames.push_back("membrane_fast_sodium_current_conductance");                        // 14.838
        mParameterMetanames.push_back("membrane_L_type_calcium_current_conductance");                        // 0.000175
        mParameterMetanames.push_back("membrane_inward_rectifier_potassium_current_conductance");         // 5.405
        mParameterMetanames.push_back("membrane_rapid_delayed_rectifier_potassium_current_conductance");    // 0.096
        mParameterMetanames.push_back("membrane_slow_delayed_rectifier_potassium_current_conductance");    // 0.245
        mParameterMetanames.push_back("membrane_sodium_calcium_exchanger_current_conductance");            // 1000
        mParameterMetanames.push_back("membrane_transient_outward_current_conductance");                    // 0.294
        mParameterMetanames.push_back("membrane_background_calcium_current_conductance");                 // 0.000592
        mParameterMetanames.push_back("membrane_background_sodium_current_conductance");                    // 0.00029
        mParameterMetanames.push_back("membrane_calcium_pump_current_conductance");                        // 0.825
        mParameterMetanames.push_back("membrane_potassium_pump_current_conductance");                    // 0.0146
        mParameterMetanames.push_back("membrane_sodium_potassium_pump_current_permeability");            // 1.362
    }
    else if (mModelChoice==6u)
    {
        mpModel.reset(new Cellohara_rudy_2011_endo_data_clampFromCellMLCvodeOpt(p_solver, p_stimulus));
        mParameterMetanames.push_back("membrane_fast_sodium_current_conductance");                        // 75
        mParameterMetanames.push_back("membrane_L_type_calcium_current_conductance");                        // 0.0001
        mParameterMetanames.push_back("membrane_background_potassium_current_conductance");               // 0.003
        mParameterMetanames.push_back("membrane_inward_rectifier_potassium_current_conductance");            // 0.1908
        mParameterMetanames.push_back("membrane_rapid_delayed_rectifier_potassium_current_conductance");    // 0.046
        mParameterMetanames.push_back("membrane_slow_delayed_rectifier_potassium_current_conductance");   // 0.0034
        mParameterMetanames.push_back("membrane_sodium_calcium_exchanger_current_conductance");           // 0.0008
        mParameterMetanames.push_back("membrane_sodium_potassium_pump_current_permeability");                // 30
        mParameterMetanames.push_back("membrane_transient_outward_current_conductance");                    // 0.02
        mParameterMetanames.push_back("membrane_background_calcium_current_conductance");                 // 2.5e-8
        mParameterMetanames.push_back("membrane_background_sodium_current_conductance");                 // 3.75e-10
        mParameterMetanames.push_back("membrane_calcium_pump_current_conductance");                      // 0.0005
        mParameterMetanames.push_back("membrane_persistent_sodium_current_conductance");                 // 0.0075
    }
    else if (mModelChoice==7u) // Davies 2012
    {
        mpModel.reset(new Celldavies_isap_2012FromCellMLCvodeDataClampOpt(p_solver, p_stimulus));
        mEndTime = 400.0;

        mParameterMetanames.push_back("membrane_fast_sodium_current_conductance");                        // 8.25
        mParameterMetanames.push_back("membrane_L_type_calcium_current_conductance");                        // 0.000243
        mParameterMetanames.push_back("membrane_inward_rectifier_potassium_current_conductance");            // 0.5
        mParameterMetanames.push_back("membrane_potassium_pump_current_conductance");                      // 0.00276
        mParameterMetanames.push_back("membrane_slow_delayed_rectifier_potassium_current_conductance");   // 0.00746925
        mParameterMetanames.push_back("membrane_rapid_delayed_rectifier_potassium_current_conductance");    // 0.0138542
        mParameterMetanames.push_back("membrane_calcium_pump_current_conductance");                        // 0.0575
        mParameterMetanames.push_back("membrane_background_calcium_current_conductance");                 // 0.0000007980336
        mParameterMetanames.push_back("membrane_sodium_calcium_exchanger_current_conductance");           // 5.85
        mParameterMetanames.push_back("membrane_sodium_potassium_pump_current_permeability");                // 0.61875
        mParameterMetanames.push_back("membrane_transient_outward_current_conductance");                    // 0.1805
        mParameterMetanames.push_back("membrane_transient_outward_chloride_current_conductance");           // 4e-7
        mParameterMetanames.push_back("membrane_background_chloride_current_conductance");                 // 0.000225
        mParameterMetanames.push_back("membrane_persistent_sodium_current_conductance");                 // 0.011
    }
    mpModel->ForceUseOfNumericalJacobian(true);

    std::cerr << "Defined parameter names vector" << std::endl << std::flush;


    if (mDog <= 0) // These protocols are only defined for synthetic data
    {
        if ( (protocolChoice==1u) || (protocolChoice==2u) || (protocolChoice==3u) ) // Just long enough for the model to produce 1 AP and settle back to a resting potential
        {
            if ( mModelChoice==1u )
            {
                mEndTime = 100.0;
            }
            else if ( mModelChoice==2u )
            {
                mEndTime = 600.0;
            }
            else if ( mModelChoice==3u )
            {
                mEndTime = 400.0;
            }
            else if ( mModelChoice==4u )
            {
                mEndTime = 400.0;
            }
            else if ( mModelChoice==5u )
            {
                mEndTime = 400.0;
            }
            else if ( mModelChoice==6u )
            {
                mEndTime = 400.0;
            }
        }
        else if ( protocolChoice==4u ) // Two pacing rates, hopefully obtain 3 APs
        {
            mEndTime = 2000;
        }
        else if (protocolChoice==5u) // 1 Hz pacing, so 2 APs, but we will set one conductance to zero (and hence block that ion current) immediately before 2nd AP
        {
            mEndTime = 2000;
        }
        else if (protocolChoice==6u) // 1 Hz pacing, so should obtain 10 APs
        {
            mEndTime = 10000;
        }
        else if (protocolChoice==7u) // 1 Hz pacing, so should obtain 20 APs
        {
            mEndTime = 20000;
        }

        if (protocolChoice==2u)
        {
            double extracell_K_conc = mpModel->GetParameter("extracellular_potassium_concentration");
            mpModel->SetParameter("extracellular_potassium_concentration", 0.5*extracell_K_conc);
        }
        else if (protocolChoice==3u)
        {
            double extracell_K_conc = mpModel->GetParameter("extracellular_potassium_concentration");
            mpModel->SetParameter("extracellular_potassium_concentration", 2*extracell_K_conc);
        }
    }


    mpModel->SetTolerances(1e-8, 1e-10);
    mpModel->SetMaxSteps(5000);

    if (mDog <= 0)
    {
        GenerateExperimentalTrace();
        AddNoiseToExperimentalTrace();
    }
    else
    {
        LoadExperimentalTrace();
    }

    // count number of times CVODE fails to solve
    mNumberOfSingularities = 0;

    std::cerr << "Leaving constructor" << std::endl << std::flush;
}



APSimulator::~APSimulator()
{
    fclose(stderr);
}

void APSimulator::SetNoiseModelParameter(const std::vector<double>& rTestParameters)
{
    mSigma = rTestParameters.back();
}

double APSimulator::GetNoiseModelParameter()
{
    return mSigma;
}

void APSimulator::AddNoiseToExperimentalTrace()
{
    RandomNumberGenerator::Instance()->Reseed(mSeed);
    double experimental_noise_sd = 0.25;
    for (unsigned i=0; i<mExperimentalTrace.size(); i++)
    {
        double random_number = RandomNumberGenerator::Instance()->StandardNormalRandomDeviate();
        mExperimentalTrace[i] +=  experimental_noise_sd*random_number;
    }
}

void APSimulator::Run(const std::vector<double>& rTestParameters)
{
    if (not mRunningToSteadyState) // want to use same ICs for each MCMC iteration, unless we're finding steady state first
    {
        if (mDog <= 0)
        {
            mpModel->SetStateVariables(mpModel->GetInitialConditions());
        }
        else
        {
            LoadStateVariables();
        }
    }

    if (mDog > 0) // Attempting to replicate the experimental conditions under which the first 200 APs were obtained
    {
        if (mModelChoice == 4u)
        {
             mpModel->SetParameter("extracellular_potassium_concentration", 5.4);
             mpModel->SetParameter("cytosolic_potassium_concentration", 130);
        }
        else if ((mModelChoice == 5u)||(mModelChoice == 6u)||(mModelChoice == 7u)) // This does, however, mean resetting these concentrations before every solve,
        {                                                                          // but we don't think it has much effect on the steady state
            mpModel->SetParameter("extracellular_potassium_concentration", 5.4);
            mpModel->SetStateVariable("cytosolic_potassium_concentration", 130);
        }
    }

    unsigned number_of_parameters = rTestParameters.size();

    // check vector of parameters is correct length, including a noise variance parameter
    if (mModelChoice==1u)  // Hodgkin-Huxley
    {
        assert(number_of_parameters==4u);
    }
    else if (mModelChoice==2u)  // Noble '62
    {
        assert(number_of_parameters==5u);
    }
    else if (mModelChoice==3u)  // Beeler-Reuter '77
    {
        assert(number_of_parameters==5u);
    }
    else if (mModelChoice==4u)  // Luo-Rudy '91
    {
        assert(number_of_parameters==7u);
    }
    else if (mModelChoice==5u)  // Ten Tusscher Epi. '04
    {
        assert(number_of_parameters==13u);
    }
    else if (mModelChoice==6u)  // O'Hara-Rudy Endo. '11
    {
        assert(number_of_parameters==14u);
    }
    else if (mModelChoice==7u)  // Davies '12
    {
        assert(number_of_parameters==15u);
    }

    // set conductances to newly proposed values before solving
    for (unsigned i=0; i<number_of_parameters-1u; i++)
    {
        mpModel->SetParameter(mParameterMetanames[i], rTestParameters[i]);
    }

    // set noise standard deviation to proposed value
    mSigma = rTestParameters.back();

    if (mSigma==DOUBLE_UNSET)
    {
        EXCEPTION("SetNoiseModelParameter() has not been called.");
    }

    // solve the model equations given proposed parameter values
    if ( (mFreezingIndex == -1) && (mDog<=0) ) // for any protocol where we don't change anything during the measurement, i.e. just let CVODE solve the system we've set up
    {
        try
        {
            OdeSolution sol = mpModel->Compute(mStartTime, mEndTime, mSamplingTimestep);
            mTestTrace = sol.GetAnyVariable("membrane_voltage");
        }
        catch (Exception &e) // in case CVODE can't solve for some reason
        {
            std::cerr << "WARNING: CVODE failed to solve with these parameters" << std::endl << std::flush;
            std::cerr << "error was " << e.GetShortMessage() << std::endl << std::flush;
            // if CVODE fails to solve, set the proposed voltage trace to 0 everywhere, which will give a very bad log-likelihood
            mTestTrace = std::vector<double>(mExperimentalTrace.size(), 0.0);
            mNumberOfSingularities++;
        }
    }
    else if ( (mProtocolChoice == 5u) && (mFreezingIndex > -1) && (mDog<=0) )
    {
        try
        {
            mEndTime_half = 1000;
            OdeSolution sol1 = mpModel->Compute(mStartTime, mEndTime_half, mSamplingTimestep);
            mTestTrace1 = sol1.GetAnyVariable("membrane_voltage");
            mpModel->SetParameter(mParameterMetanames[mFreezingIndex], 0);
            OdeSolution sol2 = mpModel->Compute(mEndTime_half, mEndTime, mSamplingTimestep);
            mTestTrace2 = sol2.GetAnyVariable("membrane_voltage");

            mTestTrace1.erase(mTestTrace1.end()-1);
            mTestTrace = mTestTrace1;
            mTestTrace.insert( mTestTrace.end(), mTestTrace2.begin(), mTestTrace2.end() );
        }
        catch (Exception &e)
        {
            std::cerr << "WARNING: CVODE failed to solve with these parameters" << std::endl << std::flush;
            std::cerr << "error was " << e.GetShortMessage() << std::endl << std::flush;
            mTestTrace = std::vector<double>(mExperimentalTrace.size(), 0.0);
            mNumberOfSingularities++;
        }
    }
    else if (mDog>0)
    {
        std::cerr << "Inside data clamp solving method" << std::endl << std::flush;
        try
        {
            if (mRunningToSteadyState)
            {
                mEndTime = mStartTime + 1000.0; // To recreate the experimental conditions, we mimic 1 Hz pacing by solving for 1000 ms
            }
            else
            {
                SetInitialVoltageToData(); // To try to fit to data, we also set the initial point to the initial data point
            }

            boost::static_pointer_cast<AbstractCvodeCellWithDataClamp>(mpModel)->TurnOffDataClamp();


            OdeSolution sol1 = mpModel->Compute(mStartTime, mDataClampOn, mSamplingTimestep);
            mTestTrace = sol1.GetAnyVariable("membrane_voltage");
            std::cerr << "Solved first bit" << std::endl << std::flush;
            std::cerr << "mTestTrace[0] = " << mTestTrace[0] << std::endl << std::flush;


            mpModel->ResetSolver();
            boost::static_pointer_cast<AbstractCvodeCellWithDataClamp>(mpModel)->TurnOnDataClamp(200);
            OdeSolution sol2 = mpModel->Compute(mDataClampOn, mDataClampOff, mSamplingTimestep);
            std::cerr << "Solved second bit" << std::endl << std::flush;

            mpModel->ResetSolver();
            boost::static_pointer_cast<AbstractCvodeCellWithDataClamp>(mpModel)->TurnOffDataClamp();
            OdeSolution sol3 = mpModel->Compute(mDataClampOff, mEndTime, mSamplingTimestep);
            std::cerr << "Solved third bit" << std::endl << std::flush;

            std::vector<double> voltage_trace_part_2 = sol2.GetAnyVariable("membrane_voltage");
            std::vector<double> voltage_trace_part_3 = sol3.GetAnyVariable("membrane_voltage");

            mTestTrace.erase(mTestTrace.end()-1);
            mTestTrace.insert( mTestTrace.end(), voltage_trace_part_2.begin(), voltage_trace_part_2.end() );
            mTestTrace.erase(mTestTrace.end()-1);
            mTestTrace.insert( mTestTrace.end(), voltage_trace_part_3.begin(), voltage_trace_part_3.end() );



        }
        catch (Exception &e)
        {
            std::cerr << "WARNING: CVODE failed to solve with these parameters" << std::endl << std::flush;
            std::cerr << "error was " << e.GetShortMessage() << std::endl << std::flush;
            mTestTrace = std::vector<double>(mExperimentalTrace.size(), 0.0);
            mNumberOfSingularities++;
        }
    }
}


double APSimulator::CalculateLogLikelihood()
{
    assert(mTestTrace.size()==mExperimentalTrace.size());

    unsigned number_of_points = mTestTrace.size();

    double square_diff = 0.0;
    for (unsigned i=0; i<number_of_points; i++)
    {
        double temp = mExperimentalTrace[i]-mTestTrace[i];
        square_diff += temp*temp;
    }
    return -square_diff/(2*mSigma*mSigma) - number_of_points*log(mSigma);
}

std::vector<double> APSimulator::ReturnTestSolution()
{
    return mTestTrace;
}

std::vector<double> APSimulator::ReturnTestTimes()
{
    return mExperimentalTimes;
}

unsigned APSimulator::ReturnNumberOfSingularities()
{
    return mNumberOfSingularities;
}

std::vector<double> APSimulator::ReturnExperimentalTrace()
{
    return mExperimentalTrace;
}

void APSimulator::GenerateExperimentalTrace()
{
    assert(mDog <= 0);

    mStartTime = 0.0;
    mSamplingTimestep = 0.2;

    std::cerr << "About to try and solve..." << std::endl << std::flush;

    // Solve model using original parameter values as reference trace

    if ( mFreezingIndex < 0 )
    {
        std::cerr << "REALLY about to try and solve..." << std::endl << std::flush;
        OdeSolution sol = mpModel->Compute(mStartTime, mEndTime, mSamplingTimestep);
        mExperimentalTrace = sol.GetAnyVariable("membrane_voltage");
        mExperimentalTimes = sol.rGetTimes();
        std::cerr << "Just solved." << std::endl << std::flush;
    }
    else if ( (mProtocolChoice == 5u) && (mFreezingIndex >= 0) )
    {
        mEndTime_half = 1000;
        OdeSolution sol1 = mpModel->Compute(mStartTime, mEndTime_half, mSamplingTimestep);
        mExptTrace1 = sol1.GetAnyVariable("membrane_voltage");
        mpModel->SetParameter(mParameterMetanames[mFreezingIndex], 0);
        OdeSolution sol2 = mpModel->Compute(mEndTime_half, mEndTime, mSamplingTimestep);
        mExptTrace2 = sol2.GetAnyVariable("membrane_voltage");

        mExptTimes1 = sol1.rGetTimes();
        mExptTimes2 = sol2.rGetTimes();
        mExptTimes1.erase(mExptTimes1.end()-1);
        mExperimentalTimes.reserve(mExptTimes1.size()+mExptTimes2.size());
        mExperimentalTimes.insert( mExperimentalTimes.end(), mExptTimes1.begin(), mExptTimes1.end() );
        mExperimentalTimes.insert( mExperimentalTimes.end(), mExptTimes2.begin(), mExptTimes2.end() );

        mExptTrace1.erase(mExptTrace1.end()-1);
        mExperimentalTrace.reserve(mExptTrace1.size()+mExptTrace2.size());
        mExperimentalTrace.insert( mExperimentalTrace.end(), mExptTrace1.begin(), mExptTrace1.end() );
        mExperimentalTrace.insert( mExperimentalTrace.end(), mExptTrace2.begin(), mExptTrace2.end() );

        std::cerr << "Experimental trace size: " << mExperimentalTrace.size() << std::endl << std::flush;
    }

    std::cerr << "Solved" << std::endl << std::flush;
}

void APSimulator::ArchiveStateVariables() // We called this from TestArchiveStateVariables.hpp
{
    OutputFileHandler handler("archive", false);
    handler.SetArchiveDirectory();
    std::string archive_directory = "projects/Jmcc_UQ/input/dog_trace/SS_for_ICs/";
    boost::filesystem::create_directories(archive_directory);
    std::string arch_name = "m_"+boost::lexical_cast<std::string>(mModelChoice)+"_d_"+boost::lexical_cast<std::string>(mDog)+".arch";
    std::ofstream ofs((archive_directory+arch_name).c_str());
    boost::archive::text_oarchive output_arch(ofs);
    output_arch <<  *mpModel;
}

void APSimulator::LoadStateVariables()
{
    std::string arch_name = "m_"+boost::lexical_cast<std::string>(mModelChoice)+"_d_150.arch"; // We expected them to mostly be the same SS conditions, but can set this to be each individual trace
    std::ifstream ifs(("projects/Jmcc_UQ/input/dog_trace/SS_for_ICs/"+arch_name).c_str(), std::ios::binary);
    boost::archive::text_iarchive input_arch(ifs);
    input_arch >> *mpModel;
}

void APSimulator::RunToSteadyState()
{
    mRunningToSteadyState = true;
}

void APSimulator::LoadExperimentalTrace()
{
    std::string path_to_AP_file ("projects/Jmcc_UQ/input/dog_traces_csv/dog_AP_trace_"+boost::lexical_cast<std::string>(mDog)+".csv");
    std::ifstream fin(path_to_AP_file.c_str());
    std::string line;

    std::cerr << "Path to AP file: " << path_to_AP_file << std::endl << std::flush;
    std::cerr << "Path to AP file (c string): " << path_to_AP_file.c_str() << std::endl << std::flush;

    std::cerr << "Inside LoadExperimentalTrace" << std::endl << std::flush;


    while ( std::getline(fin,line) )
    {
        typedef boost::escaped_list_separator<char> Separator;
        typedef boost::tokenizer<Separator> Tokenizer;
        Tokenizer tokenizer(line);
        std::vector<std::string> tokens;
        for (Tokenizer::iterator iter = tokenizer.begin(); iter != tokenizer.end(); ++iter)
        {
           tokens.push_back(*iter);
        }
        assert(tokens.size() == 2);

        mExperimentalTimes.push_back(1000*boost::lexical_cast<double>(tokens[0]));
        mExperimentalTrace.push_back(1000*boost::lexical_cast<double>(tokens[1]));
    }
    fin.close();


    mStartTime = mExperimentalTimes[0]; // hopefully -0.125
    mEndTime = mExperimentalTimes.back(); // hopefully 811.625
    mSamplingTimestep = (mEndTime-mStartTime)/(mExperimentalTimes.size()-1); // hopefully 0.25

/*  mStartTime = -0.125;
    mEndTime = 811.625;
    mSamplingTimestep = 0.25;*/ // in case the sampling timestep doesn't match up

    std::cerr << "Start time: " << mStartTime << std::endl << std::flush;
    std::cerr << "End time: " << mEndTime << std::endl << std::flush;
    std::cerr << "Sampling timestep: " << mSamplingTimestep << std::endl << std::flush;

    boost::static_pointer_cast<AbstractCvodeCellWithDataClamp>(mpModel)->SetExperimentalData(mExperimentalTimes,mExperimentalTrace);
}

std::vector<double> APSimulator::GetStateVariables()
{
    return mpModel->GetStdVecStateVariables();
}

void APSimulator::SetInitialVoltageToData()
{
    mpModel->SetStateVariable("membrane_voltage", mExperimentalTrace[0]);
}

double APSimulator::GetVariable(const std::string& rVariableName)
{
    return mpModel->GetAnyVariable(rVariableName);
}

void APSimulator::SetParameter(const std::string& rParameterName, double value)
{
    mpModel->SetParameter(rParameterName, value);
}

double APSimulator::GetParameter(const std::string& rParameterName)
{
    return mpModel->GetParameter(rParameterName);
}
