# This script produces two .txt files:
# 1. List of all final states obtained by CMA-ES followed by their respective log-likelihoods; this can be loaded directly via numpy.loadtxt for MCMC, where each row is [final states, log-likelihood]
# 2. Matrix of population size, sigma0, final log-likelihood, and time taken -- for each full CMA-ES run; can also be loaded directly via numpy.loadtxt, but it's more just for interest's sake
#    if playing with how popsize and sigma0 affect the minimisation.

# There are 4 (and a 5th, optional) command line arguments:
# 1. AP model
# 2. Protocol
# 3. The random seed for C++ when adding random noise to generate synthetic data
# 4. The index of the parameter to set to 0 after 1 s when using protocol 3
# (5. Which dog AP trace to fit to)

import sys
try:
    import cma ###########
except:
    print "No cma module detected - enter \"pip install cma\" in the command line. Exiting now."
    sys.exit()

import subprocess

import numpy as np
import numpy.random as npr

import os

from time import time

import multiprocessing

start = time()

python_seed = 1

npr.seed(python_seed)  # seed random number generator so we can reproduce results

model = int(sys.argv[1]) # which AP model to use
protocol = int(sys.argv[2]) # which protocol to use
c_seed = int(sys.argv[3]) # random seed for C++ when generating synthetic data
freeze = int(sys.argv[4]) # index of which parameter to set to 0 if using protocol 3
try:
    dog = int(sys.argv[5])
except:
    dog = 0
    
# Effectively setting our prior distribution on the conductance parameters
# We allow 0.1* to 10* the original model value for synthetic data
# We allow 0 to 100* the original model value when fitting actual canine ventricular APs

import ModelSetup

number_of_parameters, true_gs, g_parameters, g_intervals = ModelSetup.ChooseModel(model,dog)
ModelSetup.CheckProtocolChoice(protocol)
ModelSetup.CheckFreeze(freeze,number_of_parameters)
directory, final_states_file_path = ModelSetup.CreateCMAESOutputFiles(model,protocol,c_seed,freeze,python_seed,dog)

# If we think we cannot infer certain parameters, i.e. all values in our prior are equally likely,
# then we can set them to the original model values by giving their indices in the lists below.
# This makes the CMA-ES minimisation parameter space a smaller dimension, thus speeding it up.
# The same option is in the MCMC Python script.
if (model==15):
    hold_params = []
elif (model==16):
    hold_params = []
else:
    hold_params = []

num_hold = len(hold_params)
number_of_parameters -= num_hold

num_processors = multiprocessing.cpu_count()
num_processes = min(num_processors-1,number_of_parameters) # counts available cores and makes one fewer processe

if (num_processes==0): # just in case only one processor is detected
    num_processes = 1

print num_processors, "processors in total found; setting up", num_processes, "processes"

# allows communication with Chaste
EXE_NAME = "./projects/Jmcc_UQ/apps/src/McmcApp"
processes = []
for _ in range(num_processes):
    processes.append(subprocess.Popen(EXE_NAME, False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE))

print "Opened processes\n"

def GiveParametersToChaste(theta,process):
    output_string_to_c = " ".join(map(str, theta)) + '\n'
    process.stdin.write(output_string_to_c)
    return None

def GetLogLikelihoodFromChaste(process):
    answer = process.stdout.readline()
    try:
        return float(answer)
    except ValueError,e:
        print "error", e
        print "Output from C++:", answer
        sys.exit()
        
# To be called at the end, will hopefully give number of singularities,
# i.e. how many times CVODE failed to solve
def returnNumberOfSingularities():
    return int(process.stdout.readline())

# Gets to the useful bit of the app!
PROVENANCE_HEADER = "This version of Chaste was compiled on:"
for process in processes:
    reading_values, looking_for_blank = False, False
    while (not reading_values):
        line = process.stdout.readline()
        if line.startswith(PROVENANCE_HEADER):
            looking_for_blank = True
        elif looking_for_blank and line.strip() == "":
            reading_values = True

# Setting up the simulators
for process in processes:
    process.stdin.write(str(model) + '\n')
    process.stdin.write(str(protocol) + '\n')
    process.stdin.write(str(c_seed) + '\n')
    process.stdin.write(str(freeze) + '\n')
    process.stdin.write(str(dog) + '\n')


    
def NEGlogLikelihood(theta):
    temp_theta = np.copy(theta) # to not mess up theta for successive iterations
    return -logLikelihood(temp_theta)

# linear scaling from unit intervals to prior intervals
"""
def unit_hypercube_to_param_intervals(an_array):
    temp_theta = np.copy(an_array)
    for i in range(number_of_parameters):
        temp_theta[i] = g_intervals[i,0] + (g_intervals[i,1] - g_intervals[i,0]) * temp_theta[i]
    return temp_theta
"""
    
# exponential-type scaling rather than linear
def unit_hypercube_to_param_intervals(an_array):
    temp_theta = np.copy(an_array)
    for i in range(number_of_parameters):
        temp_theta[i] = g_intervals[i,0] + (g_intervals[i,1] - g_intervals[i,0] + 1) ** temp_theta[i] - 1
    return temp_theta


tolfun = 1e-11 # default
tolx = 1e-4 # we don't expect to need conductance parameters to beyond 4 sig figs


starting_point = 0.5 * np.ones(number_of_parameters)
default_popsize = 4+int(3*np.log(number_of_parameters))
print "default_popsize =", default_popsize
popsize_test_file = directory+'popsize_times_python_seed_'+str(python_seed)+'_laxer_tolerances.txt'
print popsize_test_file
with open(popsize_test_file,'w') as outfile:
    outfile.write('# Starting in the middle.\n')
    outfile.write('# popsize, sigma0, logLikelihood, time taken (s).\n')
H = []
t=0

# Generally a bit easier to reproduce results using a numpy.random.RandomState
generator = npr.RandomState(python_seed)

print number_of_parameters

# First give Chaste the original model parameters and see what log-likelihood we get
GiveParametersToChaste(true_gs,processes[0])
true_gs_ll = GetLogLikelihoodFromChaste(processes[0])

print "true_gs:", true_gs
print "true_gs_ll:", true_gs_ll

with open(final_states_file_path, 'w') as outfile:
    np.savetxt(outfile,[np.concatenate((true_gs,[true_gs_ll]))])
with open(final_states_file_path,'a') as outfile:
    outfile.write("# Original model parameters and associated log-likelihood\n")

how_many_runs = 21
for t in range(how_many_runs):
    print t+1, "/", how_many_runs
    

    if (t==0):
        starting_point = np.ones(number_of_parameters) / 100.
        sigma0 = 0.0005
    else:
        starting_point = generator.rand(number_of_parameters)
        sigma0 = 0.2
    new_start = time()
    
    ll_path = []
    means = []
    opts = cma.CMAOptions()
    popsize = default_popsize * 2
    opts['seed'] = 1
    
    opts['tolfun'] = tolfun
    opts['tolx'] = tolx
    print "Initial state:", starting_point
    print "popsize =", popsize, "- sigma0 =", sigma0
    es = cma.CMAEvolutionStrategy(starting_point, sigma0, opts)
    while not es.stop():
        X = es.ask()
        num_sets = len(X)
        logLikelihoods = [0]*num_sets
        indices_to_compute = range(num_sets)
        for pset in range(num_sets):
            for param in X[pset]:
                if param < 0:
                    logLikelihoods[pset] = 1e9 + 1e9*(0 - param) # to give bad answer plus big gradient for theta outside good interval, they are positive as we make them negative later
                    indices_to_compute.remove(pset)
                    break
                elif param > 1:
                    logLikelihoods[pset] = 1e9 + 1e9*(param - 1) # same as above
                    indices_to_compute.remove(pset)
                    break
        num_sets_to_compute = len(indices_to_compute)
        num_blocks = num_sets_to_compute / num_processes + 1

        for block in range(num_blocks):
            for i in range(num_processes):
                if (block*num_processes+i < num_sets_to_compute):
                    long_params = np.copy(X[indices_to_compute[block*num_processes+i]])
                    for index in hold_params:
                        long_params = np.insert(long_params,index,1./100)
                    GiveParametersToChaste(unit_hypercube_to_param_intervals(long_params),processes[i])
                else:
                    break
            for j in range(num_processes):
                if (block*num_processes+j < num_sets_to_compute):
                    logLikelihoods[indices_to_compute[block*num_processes+j]] = -GetLogLikelihoodFromChaste(processes[j])
                else:
                    break
        es.tell(X, logLikelihoods)
        es.disp()
    print('\nterminated on ' + str(es.stop())+'\n')
    res = es.result()
    final_state = res[0]
    print "final_state =", final_state
    for index in hold_params:
        final_state = np.insert(final_state,index,1./100)
    print "final_state =", final_state
    H.append([popsize,sigma0,-res[1],time()-new_start])
    with open(final_states_file_path,'a') as outfile:
	np.savetxt(outfile,[np.concatenate((unit_hypercube_to_param_intervals(final_state),[-res[1]]))])
    with open(final_states_file_path,'a') as outfile:
        outfile.write("# sigma0: "+str(sigma0)+", Time taken: " + str(time()-new_start) + " s\n")

with open(final_states_file_path,'a') as outfile:
    outfile.write("#\n# Total time taken: " + str(time()-start) + " s\n#\n#\n")

    print "****\nTime taken:", time()-new_start,"s\n*****"

process.stdin.write("finish\n")
H = np.array(H)

with open(popsize_test_file,'w') as outfile:
    np.savetxt(outfile, H)


print "Number of singularities:", returnNumberOfSingularities()

print "Time taken:", time()-start, "s"
