# This scripts produces one massive .txt file, where each row is one iteration of the MCMC
# Parameters, log-likelihood, acceptance rate, log(a), mean estimate, real time taken in s, iteration number.
# a, as in log(a), is the scaling factor for our Normal proposal distribution
# a will shrink if not many points are being accepted by the MCMC, and will grow if too many points are being accepted,
# with the goal of obtaining an acceptance rate of 0.25

# There are 4 (and a 5th, optional) command line arguments:
# 1. AP model
# 2. Protocol
# 3. The random seed for C++ when adding random noise to generate synthetic data
# 4. The index of the parameter to set to 0 after 1 s when using protocol 3
# (5. Which dog AP trace to fit to)

import subprocess

import numpy as np
import numpy.random as npr
import sys
import os

from time import time

start = time()

python_seed = 1

npr.seed(python_seed)  # seed random number generator so we can reproduce results

### logPrior does not appear in MCMC when assuming uniform prior on theta

# allows communication with Chaste
EXE_NAME = "./projects/Jmcc_UQ/apps/src/McmcApp"
process = subprocess.Popen(EXE_NAME, False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)

# don't need to calculate prior since it it constant in allowed region, and we ignore samples from disallowed region
def logLikelihood(theta):
    global process
    # Call out to the C++

    output_string_to_c = " ".join(map(str, theta)) + '\n'

    # give vector of parameters to C++
    process.stdin.write(output_string_to_c)
    
    # wait for output from C++
    answer = process.stdout.readline()
    try:
        return float(answer)
    except ValueError,e:
        print "error", e
        print "Output from C++:", answer
        print "theta given to C++:", theta

        
# to be called at the end, will hopefully give number of singularities; worth being a function at all?
def returnNumberOfSingularities():
    return int(process.stdout.readline()) 
    	
# prior distribution of parameters - uniform on certain intervals


PROVENANCE_HEADER = "This version of Chaste was compiled on:"
reading_values, looking_for_blank = False, False
while (not reading_values):
    line = process.stdout.readline()
    if line.startswith(PROVENANCE_HEADER):
        looking_for_blank = True
    elif looking_for_blank and line.strip() == "":
        reading_values = True

model = int(sys.argv[1])
protocol = int(sys.argv[2])
c_seed = int(sys.argv[3])
freeze = int(sys.argv[4])
try:
    dog = int(sys.argv[5])
except:
    dog = 0

    
thinning = 10 # remove auto-correlation by thinning the samples, i.e. keep every 10th step, say


process.stdin.write(str(model) + '\n')
process.stdin.write(str(protocol) + '\n')
process.stdin.write(str(c_seed) + '\n')
process.stdin.write(str(freeze) + '\n')
process.stdin.write(str(dog) + '\n')

import ModelSetup

number_of_parameters, true_gs, g_parameters, g_intervals = ModelSetup.ChooseModel(model,dog)
ModelSetup.CheckProtocolChoice(protocol)
ModelSetup.CheckFreeze(freeze,number_of_parameters)
CMAES_directory, CMAES_final_states_file = ModelSetup.CreateCMAESOutputFiles(model,protocol,c_seed,freeze,python_seed,dog)
directory, file_path, covariance_estimate_path = ModelSetup.CreateMCMCOutputFiles(model,protocol,c_seed,freeze,python_seed,dog,thinning)

CMAES_final_states = np.loadtxt(CMAES_final_states_file)
logLikelihoods = CMAES_final_states[:,-1]
max_logLikelihood_index = np.argmax(logLikelihoods)
thetaCur = CMAES_final_states[max_logLikelihood_index,:-1]
logTargetCur = logLikelihood(thetaCur) # instead of reading logTargetCur from CMA-ES file; probably safer
#assert(logTargetCur==CMAES_final_states[max_logLikelihood_index,-1]) # might not work due to rounding errors

# Same as in CMA-ES_python.py; parameters we don't think we can infer can be set to their original model values
# by giving their indices here
if (model==15):
    hold_params = []
elif (model==16):
    hold_params = []
else:
    hold_params = []
    
num_hold = len(hold_params)

indices_for_MCMC = list(set(range(number_of_parameters))-set(hold_params))
num_free = len(indices_for_MCMC)
assert(num_free == number_of_parameters-num_hold)


print "thetaCur =", thetaCur
print "logTargetCur =", logTargetCur

short_thetaCur = np.copy(thetaCur[indices_for_MCMC])
short_thetaStar = np.array([0.]*num_free)

loga = 0.
acceptance = 0.

mean_estimate = np.copy(short_thetaCur)

cov_estimate = np.identity(number_of_parameters)
for i in range(number_of_parameters):
    cov_estimate[i,i] = (0.5*(g_intervals[i,1]-g_intervals[i,0]))
cov_estimate = np.delete(cov_estimate,hold_params,axis=0)
cov_estimate = np.delete(cov_estimate,hold_params,axis=1)
assert(len(cov_estimate)==num_free)
    
print "covariance matrix:\n", cov_estimate

if not dog:
    with open(file_path, 'w') as outfile:
        outfile.write('# Model '+str(model)+', protocol ' + str(protocol) + '\n')
        outfile.write('# Adaptive covariance matrix\n')
        outfile.write('# Thinned by taking saving only every ' + str(thinning) + '-th accepted state\n')
        outfile.write('# 1-'+str(number_of_parameters)+'. parameters, '+str(1+number_of_parameters)+'. logLikelihood, '
                      +str(2+number_of_parameters)+'. acceptance rate, '+str(3+number_of_parameters)+'. loga, '
                      +str(4+1*number_of_parameters)+'-'+str(4+2*number_of_parameters-1)+'. mean estimate, '
                      +str(4+2*number_of_parameters)+'. real time taken (s), '
                      +str(5+2*number_of_parameters)+'. iteration\n')
else:
    with open(file_path, 'w') as outfile:
        outfile.write('# Dog trace ' + str(dog) + '\n')
        outfile.write('# Model '+str(model)+', protocol ' + str(protocol) + '\n')
        outfile.write('# Adaptive covariance matrix\n')
        outfile.write('# Thinned by taking saving only every ' + str(thinning) + '-th accepted state\n')
        outfile.write('# 1-'+str(number_of_parameters)+'. parameters, '+str(1+number_of_parameters)+'. logLikelihood, '
                      +str(2+number_of_parameters)+'. acceptance rate, '+str(3+number_of_parameters)+'. loga, '
                      +str(4+1*number_of_parameters)+'-'+str(4+2*number_of_parameters-1)+'. mean estimate, '
                      +str(4+2*number_of_parameters)+'. real time taken (s), '
                      +str(5+2*number_of_parameters)+'. iteration\n')

continue_forever = True
t=0

long_mean_estimate = np.copy(mean_estimate)
for i in hold_params:
    long_mean_estimate = np.insert(long_mean_estimate,i,true_gs[i])

H = [np.concatenate((np.copy(thetaCur),[logTargetCur, acceptance, loga],long_mean_estimate,[time()-start,t]))]

# Do the actual MCMC
# 1,000,000 iterations seems to normally be enough to give good convergence, but occasionally it might require longer
how_many_iterations = 1000000
while (t <= how_many_iterations):
    if t%(500)==0 and t>0:
        with open(file_path,'a') as outfile:
            np.savetxt(outfile, H)
        with open(covariance_estimate_path, 'w') as covfile:
            np.savetxt(covfile, cov_estimate)
        H = []
        if ( freeze < 0 ) and (not dog):
            print "Model " + str(model) + ", protocol " + str(protocol)
        elif (not dog):
            print "Model " + str(model) + ", protocol " + str(protocol), "freezing", g_parameters[freeze]
        else:
            print "Model " + str(model) + ", protocol " + str(protocol) + ", dog trace "+str(dog)
        print "Iteration:", t
        print "acceptance =", acceptance
        print "loga =", loga
        print "covariance of g_{Na}:", cov_estimate[0,0]
        print "Current logLikelihood:", logTargetCur
        print "Time taken:", time() - start, "sec"
        print "Proposed theta:", thetaStar
    # propose new parameter values by sampling from normal centred on current guess
    cov_matrix_for_proposal = np.exp(loga)*cov_estimate
    short_thetaStar = npr.multivariate_normal(short_thetaCur,cov_matrix_for_proposal)
    thetaStar = np.copy(short_thetaStar)
    for i in hold_params:
        thetaStar = np.insert(thetaStar,i,true_gs[i])
            
    valid = True
    for i in indices_for_MCMC:
        if (g_intervals[i,0] < thetaStar[i] < g_intervals[i,1]):
            pass
        else:
            valid = False
            break
    accepted = 0      
    if valid:
        logTargetStar = logLikelihood(thetaStar)
        u = npr.rand() # Metropolis-Hastings accept/reject step
        logu = np.log(u)
        logAlpha = logTargetStar - logTargetCur
        if logu < logAlpha:
            accepted = 1
            short_thetaCur = np.copy(short_thetaStar)
            thetaCur = np.copy(thetaStar)
            logTargetCur = logTargetStar
    
    # start to update covariance matrix
    if t > 1000*number_of_parameters:
        s = t - 1000*number_of_parameters
        gamma_s = 1/(s+1)**0.6
        temp_covariance_bit = np.array([short_thetaCur-mean_estimate])
        cov_estimate = (1-gamma_s) * cov_estimate + gamma_s * np.dot(np.transpose(temp_covariance_bit),temp_covariance_bit)
        mean_estimate = (1-gamma_s) * mean_estimate + gamma_s * short_thetaCur
        loga += gamma_s*(accepted-0.25)
    
    acceptance = (t*acceptance + accepted)/(t+1)
    
    if t%thinning==0 and t>0:
        long_mean_estimate = np.copy(mean_estimate)
        for i in hold_params:
            long_mean_estimate = np.insert(long_mean_estimate,i,true_gs[i])
        H.append(np.concatenate((np.copy(thetaCur),[logTargetCur, acceptance, loga],long_mean_estimate,[time()-start,t])))
    t=t+1

process.stdin.write("finish\n")


print "100 / 100 - Written to file."
print "Acceptance rate =", acceptance

print "Number of singularities:", returnNumberOfSingularities()

print "Time taken:", time()-start, "s"
