# This will be imported into CMA-ES and MCMC scripts after command line arguments have been read

"""
CHOICE OF CELL MODEL 

1. Hodgkin-Huxley '52
2. Noble '62
3. Beeler-Reuter '77
4. Luo-Rudy '91
5. Ten-Tusscher Epi. '04
6. O'Hara-Rudy Endo. '11
7. Davies 2012

Data clamp only available for models 3-7.
"""

"""
CHOICE OF PROTOCOL
1. 1 pacing rate
2. Extracellular potassium concentration halved
3. Extracellular potassium concentration doubled
4. 2 pacing rates
5. 1 Hz pacing for 2000 ms, with one current knocked out half-way
6. 1 Hz pacing for 10 s, i.e. 10 APs, hopefully
7. 1 Hz pacing for 20 s, i.e. 10 APs, hopefully
"""

import numpy as np
import os
import sys

how_many_models = 7
how_many_protocols = 7
sigma = 0.25 # noise standard deviation

def ChooseModel(model,dog=None):
    global g_parameters
    if (not (1 <= model <= how_many_models)):
        print "Only", how_many_models, "models available - got", model
        sys.exit()
    if model==1: # Hodgkin Huxley
        true_gs = [120,36,0.3]
        g_parameters = ['G_{Na}', 'G_K', 'G_l']
    elif model==2: # Noble
        true_gs = [400, 0.015, 1.2, 0.075]
        g_parameters = ['G_{Na}', 'G_{K1}', 'G_K', 'G_l']
    elif (model==3): # Beeler Reuter
        true_gs = [0.04, 0.0035, 0.008, 9e-4]
        g_parameters = ['G_{Na}', 'G_{K1}', 'G_K', 'G_{CaL}']
    elif (model==4): # Luo Rudy
        true_gs = [23, 0.282, 0.6047, 0.09, 0.03921, 0.0183]
        g_parameters = ['G_{Na}', 'G_K', 'G_{K1}','G_{CaL}', 'G_l', 'G_{Kp}']
    elif (model==5): # Ten Tusscher 04
        true_gs = [14.838, 0.000175, 5.405, 0.096, 0.245, 1000, 0.294, 0.000592, 0.00029, 0.825, 0.0146, 1.362]
        g_parameters = ['G_{Na}', 'G_{CaL}', 'G_{K1}', 'G_{Kr}', 'G_{Ks}', 'k_{NaCa}',
                          'G_{to}', 'G_{bCa}', 'G_{bNa}', 'G_{pCa}', 'G_{pK}', 'P_{NaK}']
    elif (model==6): # O'Hara Rudy
        true_gs = [75, 0.0001, 0.003, 0.1908, 0.046, 0.0034, 0.0008, 30, 0.02, 2.5e-8, 3.75e-10, 0.0005, 0.0075]
        g_parameters = ['G_{Na}', 'G_{CaL}', 'G_{bK}', 'G_{K1}',
                          'G_{Kr}', 'G_{Ks}', 'k_{NaCa}', 'P_{NaK}',
                          'G_{to}', 'G_{bCa}', 'G_{bNa}', 'G_{pCa}', 'G_{NaL}']
    elif (model==7): # Davies 12
        true_gs = [8.25, 0.000243, 0.5, 0.00276,
                   0.00746925, 0.0138542, 0.0575, 0.0000007980336,
                   5.85, 0.61875, 0.1805, 4e-7,
                   0.000225, 0.011]
        g_parameters = ['G_{Na}', 'G_{CaL}', 'G_{K1}', 'G_{pK}',
                        'G_{Ks}', 'G_{Kr}', 'G_{pCa}', 'G_{bCa}',
                        'k_{NaCa}', 'P_{NaK}', 'G_{to1}', 'G_{to2}',
                        'G_{bCl}', 'G_{NaL}']
                        
    assert(len(true_gs)==len(g_parameters))
    number_of_parameters = len(true_gs)+1
    true_gs.append(sigma)
    true_gs = np.array(true_gs)
    g_intervals = np.zeros((number_of_parameters,2))

    if (not dog):
        g_intervals[:,0] = 0.1*true_gs
        g_intervals[:,1] = 10*true_gs
    else:
        g_intervals[:,0] = 0
        g_intervals[:,1] = 100*true_gs

    return number_of_parameters, true_gs, g_parameters, g_intervals

def CheckProtocolChoice(protocol):
    if (not (1 <= protocol <= how_many_protocols)):
        print "Only", how_many_protocols, "protocols available, got", protocol
        sys.exit()
    else:
        return None

def CheckFreeze(freeze,number_of_parameters):
    if (freeze < 0) or (freeze > number_of_parameters-2):
        return -1
    else:
        return freeze

def CreateCMAESOutputFiles(model,protocol,c_seed,freeze,python_seed,dog=None):
    if (not dog):
        directory = 'projects/Jmcc_UQ/output/CMA-ES/synthetic_data/model_'+str(model)+'/protocol_'+str(protocol)+'/'
    else:
        directory = 'projects/Jmcc_UQ/output/CMA-ES/dog_trace/model_'+str(model)+'/dog_trace_'+str(dog)+'/'
    if (protocol==5):
        directory += g_parameters[freeze]+'/'
    directory = os.path.expanduser(directory)
    if not os.path.exists(directory):
        os.makedirs(directory)
    if (not dog):
        final_states_file_path = directory+'minimizer_final_states_python_seed_'+str(python_seed)+'_c_seed_'+str(c_seed)+'.txt'
    elif (dog):
        final_states_file_path = directory+'minimizer_final_states_python_seed_'+str(python_seed)+'_dog_trace_'+str(dog)+'.txt'
    return directory, final_states_file_path


def CreateMCMCOutputFiles(model,protocol,c_seed,freeze,python_seed,dog,thinning):
    if (not dog):
        if (protocol != 5):
            directory = 'projects/Jmcc_UQ/output/adaptive_MCMC/synthetic_data/thinned_every_'+str(thinning)+'/model_'+str(model)+'/protocol_'+str(protocol)+'/'
        else:
            directory = 'projects/Jmcc_UQ/output/adaptive_MCMC/synthetic_data/thinned_every_'+str(thinning)+'/model_'+str(model)+'/protocol_3/'+g_parameters[freeze]+'/'
    elif (dog):
        directory = 'projects/Jmcc_UQ/output/adaptive_MCMC/dog_trace/thinned_every_'+str(thinning)+'/model_'+str(model)+'/dog_trace_'+str(dog)+'/'
        
    directory = os.path.expanduser(directory)
    if not os.path.exists(directory):
        os.makedirs(directory)
    if (freeze < 0) and (protocol != 5):
        if (not dog):
            file_path = directory+'model_'+str(model)+'_protocol_'+str(protocol)+'_MCMC_noisy_adaptive_covariance_matrix_thinned_every_'+str(thinning)+'_python_seed_'+str(python_seed)+'_c_seed_'+str(c_seed)+'.txt'
            covariance_estimate_path = directory+'model_'+str(model)+'_protocol_'+str(protocol)+'_MCMC_noisy_adaptive_covariance_matrix_thinned_every_'+str(thinning)+'_python_seed_'+str(python_seed)+'_c_seed_'+str(c_seed)+'_covariance_estimate.txt'
        elif (dog):
            file_path = directory+'model_'+str(model)+'_protocol_'+str(protocol)+'_MCMC_noisy_adaptive_covariance_matrix_thinned_every_'+str(thinning)+'_python_seed_'+str(python_seed)+'_dog_trace_'+str(dog)+'.txt'
            covariance_estimate_path = directory+'model_'+str(model)+'_protocol_'+str(protocol)+'_MCMC_noisy_adaptive_covariance_matrix_thinned_every_'+str(thinning)+'_python_seed_'+str(python_seed)+'_dog_trace_'+str(dog)+'_covariance_estimate.txt'
    elif (freeze < 0) and (protocol == 3):
        file_path = directory+'model_'+str(model)+'_protocol_'+str(protocol)+'_MCMC_noisy_adaptive_covariance_matrix_thinned_every_'+str(thinning)+'_no_freezing_python_seed_'+str(python_seed)+'_c_seed_'+str(c_seed)+'.txt'
        covariance_estimate_path = directory+'model_'+str(model)+'_protocol_'+str(protocol)+'_MCMC_noisy_adaptive_covariance_matrix_thinned_every_'+str(thinning)+'_no_freezing_python_seed_'+str(python_seed)+'_c_seed_'+str(c_seed)+'_covariance_estimate.txt'
    else:
        file_path = directory+'model_'+str(model)+'_protocol_'+str(protocol)+'_MCMC_noisy_adaptive_covariance_matrix_thinned_every_'+str(thinning)+'_'+g_parameters[freeze]+'_python_seed_'+str(python_seed)+'_c_seed_'+str(c_seed)+'.txt'
        covariance_estimate_path = directory+'model_'+str(model)+'_protocol_'+str(protocol)+'_MCMC_noisy_adaptive_covariance_matrix_thinned_every_'+str(thinning)+'_'+g_parameters[freeze]+'_python_seed_'+str(python_seed)+'_c_seed_'+str(c_seed)+'_covariance_estimate.txt'
    return directory, file_path, covariance_estimate_path


