/*

Copyright (c) 2005-2015, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef TESTARCHIVESTATEVARIABLES_HPP_
#define TESTARCHIVESTATEVARIABLES_HPP_

#include <iostream>
#include "CheckpointArchiveTypes.hpp"
#include "ArchiveLocationInfo.hpp"

#include "APSimulator.hpp"

#include "BoostFilesystem.hpp"
#include "SmartPointers.hpp"

#include "HeartConfig.hpp"
#include "RegularStimulus.hpp"
#include "SimpleStimulus.hpp"
#include "ZeroStimulus.hpp"
#include "EulerIvpOdeSolver.hpp"
#include "RunAndCheckIonicModels.hpp"
#include "OdeSystemInformation.hpp"
#include "VectorHelperFunctions.hpp"

#include <CommandLineArguments.hpp>

#include <Timer.hpp>

#include <boost/lexical_cast.hpp>

//This test is always run sequentially (never in parallel)
#include "FakePetscSetup.hpp"

/* This test takes two command line arguments, each followed by an integer:
 * --dog-trace - which dog trace to clamp the stimulus current portion to
 * --model - which AP model to use.
 * The binary file has to be built and run from the command line - details on the Chaste build guide.
 *
 * We use this test to run a cell model to steady state, given 1 Hz pacing mimicked by data clamping to the stimulus current portion of a dog AP trace
 * We only used dog trace 150, since we don't think the steady states will vary much, but it can be done for any trace, and the LoadArchivedStateVariables
 * method will have to be modified to load another trace.
 */

class TestArchiveStateVariables : public CxxTest::TestSuite
{
public:
    void TestArchivingCvodeCells() throw(Exception)
    {
#ifdef CHASTE_CVODE
        //Archive

        TS_ASSERT(CommandLineArguments::Instance()->OptionExists("--dog-trace"));
        TS_ASSERT(CommandLineArguments::Instance()->OptionExists("--model"));
        unsigned dog = CommandLineArguments::Instance()->GetUnsignedCorrespondingToOption("--dog-trace");
        unsigned model_choice = CommandLineArguments::Instance()->GetUnsignedCorrespondingToOption("--model");


        const unsigned protocol_choice = 8u;
        const unsigned c_seed = 1u;
        const int freezing_index = -1;


        std::vector<double> state_vars1;
        std::vector<double> state_vars2;

        std::cout << "Model " << model_choice << ", dog " << dog << std::endl << std::flush;

        std::vector<double> true_parameters;
        if (model_choice==3u)
        {
            true_parameters.push_back(0.04);
            true_parameters.push_back(0.0035);
            true_parameters.push_back(0.008);
            true_parameters.push_back(9e-4);
        }
        else if (model_choice==4u)
        {
            true_parameters.push_back(23);
            true_parameters.push_back(0.282);
            true_parameters.push_back(0.6047);
            true_parameters.push_back(0.09);
            true_parameters.push_back(0.03921);
            true_parameters.push_back(0.0183);
        }
        else if (model_choice==5u)
        {
            true_parameters.push_back(14.838);
            true_parameters.push_back(0.000175);
            true_parameters.push_back(5.405);
            true_parameters.push_back(0.096);
            true_parameters.push_back(0.245);
            true_parameters.push_back(1000);
            true_parameters.push_back(0.294);
            true_parameters.push_back(0.000592);
            true_parameters.push_back(0.00029);
            true_parameters.push_back(0.825);
            true_parameters.push_back(0.0146);
            true_parameters.push_back(1.362);
        }
        else if (model_choice==6u)
        {
            true_parameters.push_back(75);
            true_parameters.push_back(0.0001);
            true_parameters.push_back(0.003);
            true_parameters.push_back(0.1908);
            true_parameters.push_back(0.046);
            true_parameters.push_back(0.0034);
            true_parameters.push_back(0.0008);
            true_parameters.push_back(30);
            true_parameters.push_back(0.02);
            true_parameters.push_back(2.5e-8);
            true_parameters.push_back(3.75e-10);
            true_parameters.push_back(0.0005);
            true_parameters.push_back(0.0075);
        }
        else if (model_choice==7u)
        {
            true_parameters.push_back(8.25);
            true_parameters.push_back(0.000243);
            true_parameters.push_back(0.5);
            true_parameters.push_back(0.00276);
            true_parameters.push_back(0.00746925);
            true_parameters.push_back(0.0138542);
            true_parameters.push_back(0.0575);
            true_parameters.push_back(0.0000007980336);
            true_parameters.push_back(5.85);
            true_parameters.push_back(0.61875);
            true_parameters.push_back(0.1805);
            true_parameters.push_back(4e-7);
            true_parameters.push_back(0.000225);
            true_parameters.push_back(0.011);
        }
        true_parameters.push_back(0.25);

        APSimulator simulator1(model_choice, protocol_choice, c_seed, freezing_index, dog);
        simulator1.RunToSteadyState();

        std::cerr << "Running TestArchiveStateVariables - just constructed APSimulator" << std::endl << std::flush;

        Timer::Reset();

        std::vector<double> old_state_vars;
        std::vector<double> new_state_vars;
        CopyToStdVector(simulator1.GetStateVariables(),old_state_vars);

        double max_paces = 1e5;
        double temp;
        unsigned final_pace = max_paces;
        for (unsigned i=0; i<max_paces; i++)
        {
            simulator1.Run(true_parameters);

            CopyToStdVector(simulator1.GetStateVariables(),new_state_vars);

            temp = 0.0;
            for (unsigned j=0; j<old_state_vars.size(); j++)
            {
                temp += fabs(new_state_vars[j] - old_state_vars[j]);
            }

            std::cerr << "temp = " << temp << std::endl << std::flush;

            if (temp < 1e-6)
            {
                std::cout << "temp < 1e-6 after " << i+1 << " solves" << std::endl << std::flush;
                final_pace = i;
                break;
            }

            old_state_vars = new_state_vars;
        }
        std::cout << "temp = " << temp << " at the end" << std::endl << std::flush;
        std::cout << "Stopped after pace " << final_pace << " at the end" << std::endl << std::flush;

        simulator1.ArchiveStateVariables();
        std::cerr << "Done ArchiveStateVariables()" << std::endl << std::flush;
        Timer::Print("ArchiveStateVariables() done");

        APSimulator simulator2(model_choice, protocol_choice, c_seed, freezing_index, dog);

        simulator2.LoadStateVariables();
        std::cerr << "Done LoadStateVariables()" << std::endl << std::flush;
        state_vars2 = simulator2.GetStateVariables();
        TS_ASSERT_EQUALS(new_state_vars,state_vars2); // Actually, this will only pass (I assume) if clamping to dog trace 150 (this can be changed in APSimulator.cpp)

#else
        std::cout << "Cvode is not enabled.\n";
#endif // CHASTE_CVODE
    }
};


#endif /*TESTARCHIVESTATEVARIABLES_HPP_*/
